/*
 *  Aquamarine the KDE window decorator
 *
 *  Copyright (c) 2006 Dennis Kasprzyk <onestone@beryl-project.org>
 *  Copyright (c) 2006 Volker Krause <vkrause@kde.org>
 *
 *  Uses code of:
 *      Emerald window decorator (www.beryl-project.org)
 *      gtk-window-decorator (www.freedesktop.org/wiki/Software/Compiz)
 *      KWin window manager (www.kde.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#include <qglobal.h>

#include <kcmdlineargs.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kglobal.h>
#include <kwinmodule.h>
#include <klocale.h>

#include <qpoint.h>

#include "switcher.h"
#include "decorator.h"
#include "options.h"
#include "utils.h"
#include "integration.h"

#include <stdio.h>

#define GRAVITY_WEST  (1 << 0)
#define GRAVITY_EAST  (1 << 1)
#define GRAVITY_NORTH (1 << 2)
#define GRAVITY_SOUTH (1 << 3)

#define XX_MASK (1 << 12)
#define XY_MASK (1 << 13)
#define YX_MASK (1 << 14)
#define YY_MASK (1 << 15)

#define CLAMP_HORZ (1 << 0)
#define CLAMP_VERT (1 << 1)

Aquamarine::PluginManager * Aquamarine::Decorator::m_plugins = 0;
Aquamarine::Options * Aquamarine::Decorator::m_options = 0;
QSize
    Aquamarine::Decorator::m_desktop;
QPoint
    Aquamarine::Decorator::m_viewport;
QSize
    Aquamarine::Decorator::m_rootSize;

extern Time
    qt_x_time;

struct _cursor
    cursors[3][3] = {
    {C (top_left_corner), C (top_side), C (top_right_corner)},
    {C (left_side), C (left_ptr), C (right_side)},
    {C (bottom_left_corner), C (bottom_side), C (bottom_right_corner)}
};


Aquamarine::PluginManager::PluginManager (KConfig * config):
KDecorationPlugins (config)
{
    // taken from KWin
    defaultPlugin =
        (QPixmap::defaultDepth () > 8) ? "kwin3_plastik" : "kwin3_quartz";
}



Aquamarine::Decorator::Decorator (Display * dpy, Qt::HANDLE visual, Qt::HANDLE colormap):
DCOPObject ("KWinInterface"),
KApplication (dpy, visual, colormap),
m_defDeco (0),
m_defDecoActive (0),
m_activeWindow (0),
m_config (0),
m_kwinmodule (new KWinModule (this, KWinModule::INFO_ALL)), m_switcher (0)
{
    delete KGlobal::_locale;
    KGlobal::_locale = new KLocale( "kwin" , NULL );

    ready = false;
    m_restart = false;

    m_plugins = NULL;
    m_options = NULL;

    Atoms::init ();

    m_activeWM = XGetSelectionOwner(qt_xdisplay(),Atoms::wm_sn);
    if (m_activeWM == None)
        return;
    XSelectInput (qt_xdisplay(), m_activeWM, StructureNotifyMask);

    KCmdLineArgs *
        args = KCmdLineArgs::parsedArgs ();
    bool
        replace = args->isSet ("replace");
    ready = acquireDmSession (qt_xdisplay (), qt_xscreen (), replace);
    if (!ready)
        return;

    set_dm_check_hint (qt_xscreen ());

    m_config = new KConfig ("kwinrc");
    m_config->setGroup ("Style");

#ifdef HAVE_BERYL_SETTINGS
	m_integrate = args->isSet ("integration");
	printf("Integration : %s\n",m_integrate? "Enabled" : "Disabled");
	if (m_integrate)
	{
		m_integration = new Aquamarine::Integration ();
	}
#endif
	connect (this, SIGNAL (settingsChanged (int)),
             SLOT (slotSettingsChanged (int)));

    QString
        decoName;
    if (args->isSet ("deco"))
        decoName = args->getOption ("deco");
    if (!decoName.startsWith ("kwin3_") && !decoName.isEmpty ())
        decoName = "kwin3_" + decoName;

    m_options = new Aquamarine::Options (m_config);
    m_plugins = new PluginManager (m_config);
    kdDebug () << "loading plugin: " << decoName << endl;
    bool
        rv = pluginManager ()->loadPlugin (decoName);
    kdDebug () << "plugin loaded: " << rv << endl;

    updateDesktopGeometry ();

    m_defDeco =
        new Aquamarine::Window (qt_xrootwin (), 0,
                                Aquamarine::Window::DecorationDefault);
    m_defDecoActive =
        new Aquamarine::Window (qt_xrootwin (), 0,
                                Aquamarine::Window::DecorationDefaultActive);

    connect (m_kwinmodule, SIGNAL (windowAdded (WId)),
             SLOT (slotWindowAdded (WId)));
    connect (m_kwinmodule, SIGNAL (windowRemoved (WId)),
             SLOT (slotWindowRemoved (WId)));
    connect (m_kwinmodule, SIGNAL (activeWindowChanged (WId)),
             SLOT (slotActiveWindowChanged (WId)));
    connect (m_kwinmodule,
             SIGNAL (windowChanged (WId, const unsigned long *)),
             SLOT (slotWindowChanged (WId, const unsigned long *)));

    QValueList < WId >::ConstIterator it;
    for (it = m_kwinmodule->windows ().begin ();
         it != m_kwinmodule->windows ().end (); ++it)
    {
        slotWindowAdded ((*it));
    }


    for (int i = 0; i < 3; i++)
    {
        for (int j = 0; j < 3; j++)
        {
            if (cursors[i][j].shape != XC_left_ptr)
                cursors[i][j].cursor =
                    XCreateFontCursor (qt_xdisplay (), cursors[i][j].shape);
        }
    }

    connect( this, SIGNAL(appearanceChanged()), SLOT(reconfigure()) );
}

Aquamarine::Decorator::~Decorator ()
{
    QMap < WId, Aquamarine::Window * >::ConstIterator it;
    for (it = m_windows.begin ();
         it != m_windows.end (); ++it)
    {
        delete (*it);
    }
    delete
        m_defDeco;
    delete
        m_defDecoActive;
    delete
        m_options;
    delete
        m_plugins;
    delete
        m_config;
    delete
        m_switcher;
    delete
        m_kwinmodule;
    if (dm_window != None) XDestroyWindow(qt_xdisplay(),dm_window);
    dm_window = None;
}

bool
Aquamarine::Decorator::x11EventFilter (XEvent * xevent)
{
    WId
        xid = 0;


    switch (xevent->type)
    {
        case DestroyNotify:
            if (xevent->xdestroywindow.window == m_activeWM)
            {
                m_restart = true;
                KApplication::exit(0);
            }
            break;
        case SelectionRequest:
            handleSelectionRequest (xevent);
            break;
        case SelectionClear:
            if (xevent->xselectionclear.selection == Atoms::dm_sn)
            {
                m_restart = false;
                KApplication::exit(0);
            }
            break;
        case ConfigureNotify:
        {
            XConfigureEvent *
                xce = reinterpret_cast < XConfigureEvent * >(xevent);
            Aquamarine::Window * client = 0;
            if (m_frames.contains (xce->window))
                client = m_frames[xce->window];
            if (!client)
                break;
            client->updateFrame (xce->window);
        }
        case PropertyNotify:
            if (xevent->xproperty.atom == Atoms::net_frame_window)
            {
                xid = xevent->xproperty.window;
                slotWindowAdded (xid);
            }
            if (xevent->xproperty.atom == Atoms::select_window)
            {
                xid = xevent->xproperty.window;
                if (!m_switcher || m_switcher->xid () != xid)
                    slotWindowAdded (xid);
                m_switcher->update ();

            }
            if (xevent->xproperty.atom == Atoms::net_desktop_geometry
                || xevent->xproperty.atom == Atoms::net_desktop_viewport)
            {
                updateDesktopGeometry ();
            }
            if (xevent->xproperty.atom == Atoms::opacity)
            {
                Aquamarine::Window * client = 0;
                if (m_windows.contains (xevent->xproperty.window))
                    client = m_windows[xevent->xproperty.window];
                if (client)
                    client->updateOpacity ();
            }
            if (xevent->xproperty.atom == Atoms::brightness)
            {
                Aquamarine::Window * client = 0;
                if (m_windows.contains (xevent->xproperty.window))
                    client = m_windows[xevent->xproperty.window];
                if (client)
                    client->updateBrightness ();
            }
            if (xevent->xproperty.atom == Atoms::saturation)
            {
                Aquamarine::Window * client = 0;
                if (m_windows.contains (xevent->xproperty.window))
                    client = m_windows[xevent->xproperty.window];
                if (client)
                    client->updateSaturation ();
            }
            break;
        case EnterNotify:
        {
            XCrossingEvent *
                xce = reinterpret_cast < XCrossingEvent * >(xevent);
            Aquamarine::Window * client = 0;
            if (m_frames.contains (xce->window))
                client = m_frames[xce->window];
            if (!client)
                break;
            QWidget *
                widget = client->decoration ()->widget ();
            QWidget *
                child = widget->childAt (xce->x, xce->y, true);
            if (!child)
                break;

            QEvent qe (QEvent::Enter);
            QApplication::sendEvent (child, &qe);
            client->setActiveChild (child);
            client->updateCursor (QPoint (xce->x, xce->y));
            break;
        }
        case LeaveNotify:
        {
            XCrossingEvent *
                xce = reinterpret_cast < XCrossingEvent * >(xevent);
            Aquamarine::Window * client = 0;
            if (m_frames.contains (xce->window))
                client = m_frames[xce->window];
            if (!client)
                break;

            QEvent qe (QEvent::Leave);
            QApplication::sendEvent (client->activeChild (), &qe);
            XUndefineCursor (qt_xdisplay (), client->frameId ());
            break;
        }
        case MotionNotify:
        {
            XMotionEvent *
                xme = reinterpret_cast < XMotionEvent * >(xevent);
            Aquamarine::Window * client = 0;
            if (m_frames.contains (xme->window))
                client = m_frames[xme->window];
            if (!client)
                break;
            QWidget *
                widget = client->decoration ()->widget ();
            QWidget *
                child = widget->childAt (xme->x, xme->y, true);
            if (!child)
                break;
            if (child != client->activeChild ())
            {
                QEvent qee (QEvent::Enter);
                QEvent qle (QEvent::Leave);
                if (client->activeChild ())
                    QApplication::sendEvent (client->activeChild (), &qle);
                QApplication::sendEvent (child, &qee);
                client->setActiveChild (child);
            }
            QPoint qp (xme->x, xme->y);
            if (widget != child)
                qp -= QPoint (child->pos ().x (), child->pos ().y ());
            QMouseEvent qme (QEvent::MouseMove, qp,
                             Qt::NoButton, Qt::NoButton);
            QApplication::sendEvent (child, &qme);

            client->updateCursor (QPoint (xme->x, xme->y));

            break;
        }
        case ButtonPress:
        case ButtonRelease:
        {
            XButtonEvent *
                xbe = reinterpret_cast < XButtonEvent * >(xevent);
            Aquamarine::Window * client = 0;
            if (m_frames.contains (xbe->window))
                client = m_frames[xbe->window];
            if (!client)
                break;

            QWidget *
                widget = client->decoration ()->widget ();
            QWidget *
                child = widget->childAt (xbe->x, xbe->y, true);
            if (!child)
                break;

            XButtonEvent *
                xbe2 = new XButtonEvent;
            memcpy (xbe2, xbe, sizeof (XButtonEvent));
            xbe2->window = child->winId ();
            if (widget != child)
            {
                xbe2->x = xbe->x - child->pos ().x ();
                xbe2->y = xbe->y - child->pos ().y ();
            }
            QApplication::x11ProcessEvent ((XEvent *) xbe2);
            return true;
        }
        default:
            break;
    }
    return KApplication::x11EventFilter (xevent);
}

WId
Aquamarine::Decorator::fetchFrame (WId window)
{
    Atom
        type;
    int
        format;
    unsigned long
        nitems;
    unsigned long
        bytes_after;
    WId *
        w;
    WId **ww = &w;
    int
        err,
        result;

    Aquamarine::trapXError ();

    type = None;
    result = XGetWindowProperty (qt_xdisplay (),
                                 window,
                                 Atoms::net_frame_window,
                                 0, LONG_MAX,
                                 False, XA_WINDOW, &type, &format, &nitems,
                                 &bytes_after, (unsigned char **)ww);
    err = Aquamarine::popXError ();

    if (err || result != Success)
        return 0;

    if (type != XA_WINDOW)
    {
        XFree (w);
        return 0;
    }

    WId
        rv = *w;
    XFree (w);

    if (validWindow(rv))
        return rv;
    else
        return 0;
}

void
Aquamarine::Decorator::reconfigure ()
{
    kdDebug () << k_funcinfo << endl;

    m_config->reparseConfiguration ();
    unsigned long
        changed = m_options->updateSettings ();

#ifdef HAVE_BERYL_SETTINGS
	if (m_integrate)
		m_integration->updateSettings();
#endif

    if (m_plugins->reset (changed))
    {
        for (QMap < WId, Aquamarine::Window * >::ConstIterator it =
             m_windows.constBegin (); it != m_windows.constEnd (); ++it)
            it.data ()->reloadDecoration ();
        m_defDeco->reloadDecoration ();
        m_defDecoActive->reloadDecoration ();
        m_plugins->destroyPreviousPlugin ();
    }
}

void
Aquamarine::Decorator::slotSettingsChanged (int)
{
	kdDebug () << k_funcinfo << endl;
#ifdef HAVE_BERYL_SETTINGS
	if (m_integrate)
		m_integration->updateSettings();
#endif
}

void
Aquamarine::Decorator::slotWindowAdded (WId id)
{
    kdDebug () << k_funcinfo << endl;
    if (!validWindow(id)) return;
    Aquamarine::trapXError ();
    XSelectInput (qt_xdisplay (), id, defaultMask);
    if (Aquamarine::popXError ())
        return;
    if (Aquamarine::readWindowProperty (id, Atoms::select_window, NULL))
    {
        if (!m_switcher)
            m_switcher = new Switcher (id);
        if (m_switcher->xid () != id)
        {
            delete
                m_switcher;
            m_switcher = new Switcher (id);
        }
        return;
    }
    WId
        frame = fetchFrame (id);
    Aquamarine::Window * client = 0;
    for (QMap < WId, Aquamarine::Window * >::ConstIterator it =
         m_windows.constBegin (); it != m_windows.constEnd (); ++it)
    {
        if (it.data ()->winId () == id)
        {
            return;
        }
    }
    if (frame)
    {
        if (!m_windows.contains (id))
        {
            client = new Aquamarine::Window (id, frame);
            connect( this, SIGNAL(desktopGeometryChanged()), client, SLOT(slotDesktopGeometryChanged()) );
            if (m_kwinmodule->activeWindow () == id)
            {
                if (m_activeWindow)
                    m_activeWindow->setActive (false);
                m_activeWindow = client;
                client->setActive(true);
            }
            m_windows.insert (id, client);
            m_frames.insert (frame, client);
        }
        else
        {
            client = m_windows[id];
            m_frames.remove (client->frameId ());
            m_frames.insert (frame, client);
            client->updateFrame (frame);
        }
    }
    else
    {
        if (m_windows.contains (id))
        {
            client = m_windows[id];
        }
        if (client)
        {
            m_windows.remove (client->windowId ());
            m_frames.remove (client->frameId ());
            delete
                client;
        }
    }
}

void
Aquamarine::Decorator::slotWindowRemoved (WId id)
{
    kdDebug () << k_funcinfo << endl;
    Aquamarine::Window * window = 0;
    if (m_windows.contains (id))
    {
        window = m_windows[id];
    }
    else if (m_frames.contains (id))
    {
        window = m_frames[id];
    }
    if (window)
    {
        m_windows.remove (window->windowId ());
        m_frames.remove (window->frameId ());
        if (m_activeWindow == window)
            m_activeWindow = 0;
        delete
            window;
    }
    if (m_switcher && m_switcher->xid () == id)
    {
        delete
            m_switcher;
        m_switcher = NULL;
    }
}

void
Aquamarine::Decorator::slotActiveWindowChanged (WId id)
{
    kdDebug () << k_funcinfo << endl;

    Aquamarine::Window * newActiveWindow = 0;
    if (m_windows.contains (id))
        newActiveWindow = m_windows[id];

    if (newActiveWindow && newActiveWindow != m_activeWindow)
    {
        if (m_activeWindow)
        {
            Aquamarine::trapXError ();
            XWindowAttributes
                attr;
            if (XGetWindowAttributes
                (qt_xdisplay (), m_activeWindow->windowId (), &attr)
                && !Aquamarine::popXError ())
                m_activeWindow->setActive (false);
        }
        newActiveWindow->setActive (true);
    }
    m_activeWindow = newActiveWindow;
}

void
Aquamarine::Decorator::slotWindowChanged (WId id,
                                          const unsigned long *properties)
{
    Aquamarine::Window * client = 0;
    if (m_windows.contains (id))
        client = m_windows[id];

    if (properties[0] & NET::WMName)
    {
        if (client)
            client->updateName ();
    }
    if (properties[0] & NET::WMState)
    {
        if (client)
            client->updateState ();
    }
    if (properties[0] & NET::WMIcon)
    {
        if (client)
            client->updateIcons ();
    }
    if (properties[0] & NET::WMGeometry)
    {
        if (m_switcher && m_switcher->xid () == id)
            m_switcher->updateGeometry ();
        if (client)
            client->updateWindowGeometry ();
    }
}

void
Aquamarine::Decorator::sendClientMessage (WId w, Atom a, Atom protocol,
                                          long data1, long data2, long data3)
{
    XEvent
        ev;
    long
        mask;

    memset (&ev, 0, sizeof (ev));
    ev.xclient.type = ClientMessage;
    ev.xclient.window = w;
    ev.xclient.message_type = a;
    ev.xclient.format = 32;
    ev.xclient.data.l[0] = protocol;
    ev.xclient.data.l[1] = qt_x_time;
    ev.xclient.data.l[2] = data1;
    ev.xclient.data.l[3] = data2;
    ev.xclient.data.l[4] = data3;
    mask = 0L;
    if (w == qt_xrootwin ())
        mask = SubstructureRedirectMask;
    XSendEvent (qt_xdisplay (), w, False, mask, &ev);
}

void
Aquamarine::Decorator::updateDesktopGeometry ()
{
    int
        x,
        y;
    unsigned int
        width,
        height,
        border,
        depth;
    ::Window root;
    XGetGeometry (qt_xdisplay (), qt_xrootwin (), &root, &x, &y, &width,
                  &height, &border, &depth);
    m_rootSize = QSize (width, height);

    int
        num_val;
    void *
        values =
        readXProperty (qt_xrootwin (), Atoms::net_desktop_geometry,
                       XA_CARDINAL, &num_val);

    if (values)
    {
        unsigned long *
            val = (unsigned long *)values;
        m_desktop = QSize (val[0], val[1]);
        XFree (val);
    }
    else
    {
        m_desktop = QSize (m_rootSize.width (), m_rootSize.height ());
    }
    values =
        readXProperty (qt_xrootwin (), Atoms::net_desktop_viewport,
                       XA_CARDINAL, &num_val);

    if (values)
    {
        unsigned long *
            val = (unsigned long *)values;
        m_viewport = QPoint (val[0], val[1]);
        XFree (val);
    }
    else
    {
        m_viewport = QPoint (0, 0);
    }

    emit desktopGeometryChanged();
}

int
Aquamarine::Decorator::onViewport (QPoint pos)
{
    return (((pos.y () +
              m_viewport.y ()) / m_rootSize.height ()) * (m_desktop.width () /
                                                          m_rootSize.
                                                          width ())) +
        ((pos.x () + m_viewport.x ()) / m_rootSize.width ()) + 1;
}

QSize
Aquamarine::Decorator::viewports ()
{
    return QSize ((m_desktop.width () / m_rootSize.width ()),
                  (m_desktop.height () / m_rootSize.height ()));
}

void
Aquamarine::Decorator::moveToViewport (WId id, int vp)
{
    int
        x,
        y;
    unsigned int
        width,
        height,
        border,
        depth;
    ::Window root;
    XGetGeometry (qt_xdisplay (), id, &root, &x, &y, &width, &height, &border,
                  &depth);
    if (vp <= 0) vp+= (m_desktop.width () / m_rootSize.width ()) * (m_desktop.height () / m_rootSize.height ());
    vp = ((vp - 1) % ((m_desktop.width () / m_rootSize.width ()) * (m_desktop.height () / m_rootSize.height ()))) + 1;
    int
        offsetX = m_rootSize.width () * ((vp - 1) % viewports ().width ());
    int
        offsetY = m_rootSize.height () * ((vp - 1) / viewports ().width ());
    XMoveWindow (qt_xdisplay (), id, x + offsetX - m_viewport.x (),
                 y + offsetY - m_viewport.y ());
}
