/*
 *  Aquamarine the KDE window decorator
 *
 *  Copyright (c) 2006 Dennis Kasprzyk <onestone@beryl-project.org>
 *  Copyright (c) 2006 Volker Krause <vkrause@kde.org>
 *
 *  Uses code of:
 *      Emerald window decorator (www.beryl-project.org)
 *      gtk-window-decorator (www.freedesktop.org/wiki/Software/Compiz)
 *      KWin window manager (www.kde.org)
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 *
 */

#ifndef AQUAMARINE_DECORATOR_H
#define AQUAMARINE_DECORATOR_H

#ifdef HAVE_CONFIG_H
#  include "../config.h"
#endif

#include <kapplication.h>

#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/cursorfont.h>

#include <fixx11h.h>
#include <kconfig.h>
#include <kdecoration_plugins_p.h>
#include <kdecoration_p.h>

#include "window.h"
#include "KWinInterface.h"

#define C(name) { 0, XC_ ## name }

#define DECOR_INTERFACE_VERSION 20061011

struct _cursor
{
    Cursor cursor;
    unsigned int shape;
};

class KConfig;
class KWinModule;



static const long defaultMask = StructureNotifyMask | PropertyChangeMask;

namespace Aquamarine
{

    class Options;
    class Switcher;

#ifdef HAVE_BERYL_SETTINGS
	class Integration;
#endif

    class PluginManager:public KDecorationPlugins
    {
      public:
        PluginManager (KConfig * config);
        virtual bool provides (Requirement)
        {
            return false;
        }
    };


    class Decorator:public KApplication, public KWinInterface
    {
      Q_OBJECT public:
        Decorator (Display * dpy, Qt::HANDLE visual = 0, Qt::HANDLE colormap =
                   0);
        ~Decorator ();

        bool isReady ()
        {
            return ready;
        }
        bool x11EventFilter (XEvent *);

        static PluginManager *pluginManager ()
        {
            return m_plugins;
        }
        static Aquamarine::Options * options ()
        {
            return m_options;
        }

        static void sendClientMessage (WId w, Atom a, Atom protocol,
                                       long data1 = 0, long data2 =
                                       0, long data3 = 0);

        static int onViewport (QPoint pos);
        static QSize viewports ();
        static void moveToViewport (WId id, int vp);

        bool restart () const
        {
            return m_restart;
        }

      public slots:
        void reconfigure ();

      signals:
        void desktopGeometryChanged();

      private:
        WId fetchFrame (WId window);
        void updateDesktopGeometry ();

      private slots:
        void slotWindowAdded (WId id);
        void slotWindowRemoved (WId id);
        void slotActiveWindowChanged (WId id);
        void slotWindowChanged (WId id, const unsigned long *properties);
		void slotSettingsChanged (int category);

      private:

        bool ready;

        Aquamarine::Window * m_defDeco;
        Aquamarine::Window * m_defDecoActive;

        QMap < WId, Aquamarine::Window * >m_windows;
        QMap < WId, Aquamarine::Window * >m_frames;
        Aquamarine::Window * m_activeWindow;

        KConfig *m_config;

        static PluginManager *m_plugins;
        static Aquamarine::Options * m_options;

        KWinModule *m_kwinmodule;

        Switcher *m_switcher;

        static QSize m_desktop;
        static QPoint m_viewport;
        static QSize m_rootSize;

        WId m_activeWM;
        bool m_restart;

#ifdef HAVE_BERYL_SETTINGS
		bool m_integrate;
		Aquamarine::Integration * m_integration;
#endif
    };

}

#endif
